/*
 * gamelib.cpp: ɮxs䴩Cclassimplementation
 * Copyright (C) 2002-2005 Woei-Kae Chen <wkc@csie.ntut.edu.tw>
 *
 * This file is part of game, a free game development framework for windows.
 *
 * game is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * game is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Known Bugs:
 *   2005-9-29
 *      1. Audio is not recovered after a suspension of windows.
 *      2. VC++.net is unable to add new event handlers for game (this is due
 *         to incompatibility of MFC files between VC++6.0 and VC++.net).
 *         Work around: I have added MOUSEMOVE handlers by using VC++6.0. Other
 *           event handlers are likely unnecessary.
 *         
 *
 * History:
 *	 1999-09-24 V2.2
 *		1. Support playback of WAVE files with DirectSound.
 *		2. Use OnIdle to control the flow of the game.
 *	 2002-02-23 v3.0 (fullscreen)
 *		1. Support fullscreen mode.
 *		2. Support playback of both MIDI and WAVE files with DirectMusic.
 *		3. Fix surface lost bugs by restoring surfaces.
 *		4. Move CGame and CBall classes into mygame.cpp.
 *		5. Add CAnimation Class.
 *   2002-03-04 V3.1
 *      1. Add CMovingBitmap::ShowBitmap(CMovingBitmap &) to enable
 *         the operation of blitting from one bitmap into another bitmap.
 *		2. Fix a CheckDDError() bug (incorrect number of errors).
 *      3. Revise delay codes for CSpecialEffect::Delay() and
 *		   CSpecialEffect::DelayFromSetCurrentTime().
 *		4. Support AUDIO_ENABLE flag.
 *   2004-03-01 V4.0
 *      1. Enhance CAudio to support pause and resume. 
 *   2004-03-08 V4.1
 *      1. Add KeyDown & KeyUp handler and remove auto-repeat for key
 *         down.
 *      2. Fix a surface lost bug due to suspend of windows.
 *      3. The Game Engine is now closer to a framework.
 *   2005-07-28 V4.2
 *      1  GAME_ASSERT is used to enforce correctness of operations for
 *         CMovingBitamp (make sure bitmap is loaded first).
 *      2. Change the constructor of CAudio to eliminate the compiling
 *         error with VC++.net.
 *      3. Make SurfaceID unsigned to eliminate warning with VC++.net.
 *   2005-09-08
 *      1. Fix a bug that handles CLR_INVALID incorrectly during
 *         SetColorKey(), BltBitmapToBack(), and BltBitmapToBitmap().
 *      2. Eliminate the use of CSpecialEffect::Abort() from BltBitmapToBack.
 *         Use GAME_ASSERT instead.
 *   2005-09-20 V4.2Beta1.
 *   2005-09-29 V4.2Beta2.
 *      1. Add MOUSEMOVE Handler for CGame and CGameState.
 *      2. Add _TRACE preprocessor flag for VC++.net.
 *   2006-02-08 V4.2
 *      1. Fix bugs: make CAnimation::Top() and CAnimation::Left() return y and x.
 *      2. Enhance CAnimation to support SetDelayCount(), Reset(), and IsFinalBitmap().
 *      3. Remove CAnimation::GetLocation() and CMovingBitmap::GetLocation().
 *      4. Bitmap coordinate can no longer be set by CMovingBitmap::LoadBitmap().
 *         Defauts to (0,0).
*/

#define	 INITGUID
#include "stdafx.h"
#include "game.h"
#include "MainFrm.h"
#include "Resource.h"
#include <mmsystem.h>
#include <ddraw.h>
#include <dsound.h>
#include <direct.h>
#include <string.h>
#include <dmusici.h>
#include "gamelib.h"
#include "mygame.h"

/////////////////////////////////////////////////////////////////////////////
// CAnimation: ѰʵeO
// 1. noIs(B)UدOAOiHUC{ON
// 2. ۤvgBCMovingBitmap{ɡAiHѦҤUC{gk
// 3. pGݭnקXRCAnimation\ɡAХ~өΪ覡A̦n
//    nCAnimationC
/////////////////////////////////////////////////////////////////////////////

CAnimation::CAnimation(int count)
{
	delay_count = count;
	delay_counter = delay_count;
	x = y = bmp_counter = 0;
}

void CAnimation::AddBitmap(int IDB_BITMAP, COLORREF colorkey) 
{
	CMovingBitmap add_bmp;
	add_bmp.LoadBitmap(IDB_BITMAP, colorkey);
	bmp.insert(bmp.end(), add_bmp);
	Reset();
}

int CAnimation::GetCurrentBitmapNumber()
{
	return bmp_counter;
}

int CAnimation::Height()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	return bmp_iter->Height();
}

bool CAnimation::IsFinalBitmap()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	return bmp_counter == (bmp.size()-1);
}

int CAnimation::Left()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	return x;
}

void CAnimation::Move()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	if (--delay_counter <= 0)  {
		delay_counter = delay_count;
		bmp_iter++;
		bmp_counter++;
		if (bmp_iter == bmp.end()) {
			bmp_iter = bmp.begin();
			bmp_counter = 0;
		}
	}
}

void CAnimation::Reset()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	delay_counter = delay_count;
	bmp_iter = bmp.begin();
	bmp_counter = 0;
}

void CAnimation::SetDelayCount(int dc)
{
	GAME_ASSERT(dc > 0, "CAnimation: Delay count must be >= 1");
	delay_count = dc;
}

void CAnimation::SetTopLeft(int nx, int ny)
{
	x = nx, y = ny;
	bmp_iter->SetTopLeft(x, y);
}

void CAnimation::Show()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded before they are shown.");
	bmp_iter->SetTopLeft(x,y);
	bmp_iter->ShowBitmap();
}

int CAnimation::Top()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	return y;
}

int CAnimation::Width()
{
	GAME_ASSERT(bmp.size() != 0,"CAnimation: Bitmaps must be loaded first.");
	return bmp_iter->Width();
}

/////////////////////////////////////////////////////////////////////////////
// CInteger: oclassܾƹϧΪO
// 1. noIs(B)UدOAOiHUC{ON
// 2. ۤvgBCMovingBitmap{ɡAiHѦҤUC{gk
/////////////////////////////////////////////////////////////////////////////

CMovingBitmap CInteger::digit[11];

CInteger::CInteger(int digits)
: NUMDIGITS(digits)
{
	isBmpLoaded = false;
}

void CInteger::Add(int x)
{
	n += x;
}

int CInteger::GetInteger()
{
	return n;
}

void CInteger::LoadBitmap()
{
	//
	// digit[i]class varibaleAҥHקKLoadBitmap
	//
	if (!isBmpLoaded) {
		int d[11]={IDB_0,IDB_1,IDB_2,IDB_3,IDB_4,IDB_5,IDB_6,IDB_7,IDB_8,IDB_9,IDB_MINUS};
		for (int i=0; i < 11; i++)
			digit[i].LoadBitmap(d[i],RGB(255,255,255));
		isBmpLoaded = true;
	}
}

void CInteger::SetInteger(int i)
{
	n = i;
}

void CInteger::SetTopLeft(int nx, int ny)		// NʵeWyв (x,y)
{
	x = nx; y = ny;
}

void CInteger::ShowBitmap()
{
	GAME_ASSERT(isBmpLoaded, "CInteger: ХLoadBitmapAM~ShowBitmap");
	int nx;		// ܦƪ x y
	int MSB;	// ̥(tŸ)ƪƭ
	if (n >= 0) {
		MSB = n;
		nx = x+digit[0].Width()*(NUMDIGITS-1);
	} else {
		MSB = -n;
		nx = x+digit[0].Width()*NUMDIGITS;
	}
	for (int i=0; i < NUMDIGITS; i++) {
		int d = MSB % 10;
		MSB /= 10;
		digit[d].SetTopLeft(nx, y);
		digit[d].ShowBitmap();
		nx -= digit[d].Width();
	}
	if (n < 0) { // pGp0Ahܭt
		digit[10].SetTopLeft(nx, y);
		digit[10].ShowBitmap();
	}
}

/////////////////////////////////////////////////////////////////////////////
// CMovingBitmap: Moving Bitmap class
// oclassѥiHʪϧ
// noIs(B)UدOAOiHUC{ON
/////////////////////////////////////////////////////////////////////////////

CMovingBitmap::CMovingBitmap()
{
	isBitmapLoaded = false;
}

int CMovingBitmap::Height()
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before Height() is called !!!");
	return location.bottom - location.top;
}

int CMovingBitmap::Left()
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before Left() is called !!!");
	return location.left;
}

void CMovingBitmap::LoadBitmap(int IDB_BITMAP, COLORREF color)
{
	const int nx = 0;
	const int ny = 0;
	GAME_ASSERT(!isBitmapLoaded,"A bitmap has been loaded. You can not load another bitmap !!!");
	CBitmap bitmap;
	BITMAP bitmapSize;
	BOOL rval = bitmap.LoadBitmap(IDB_BITMAP);
	GAME_ASSERT(rval,"Load bitmap failed !!! Please check bitmap ID (IDB_XXX).");
	bitmap.GetBitmap(&bitmapSize);
	location.left = nx; location.top = ny;
	location.right = nx+bitmapSize.bmWidth;
	location.bottom = ny+bitmapSize.bmHeight;
	SurfaceID = CDDraw::RegisterBitmap(IDB_BITMAP, color);
	isBitmapLoaded = true;
}

void CMovingBitmap::SetTopLeft(int x, int y)
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before SetTopLeft() is called !!!");
	int dx = location.left - x;
	int dy = location.top - y;
	location.left = x;
	location.top = y;
	location.right -= dx;
	location.bottom -= dy;
}

void CMovingBitmap::ShowBitmap()
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before ShowBitmap() is called !!!");
	CDDraw::BltBitmapToBack(SurfaceID,location.left,location.top);
}

void CMovingBitmap::ShowBitmap(CMovingBitmap &bm)
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before ShowBitmap() is called !!!");
	GAME_ASSERT(bm.isBitmapLoaded,"A bitmap must be loaded before ShowBitmap() is called !!!");
	CDDraw::BltBitmapToBitmap(SurfaceID, bm.SurfaceID, location.left,location.top);
}

int CMovingBitmap::Top()
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before Top() is called !!!");
	return location.top;
}

int CMovingBitmap::Width()
{
	GAME_ASSERT(isBitmapLoaded,"A bitmap must be loaded before Width() is called !!!");
	return location.right - location.left;
}

/////////////////////////////////////////////////////////////////////////////
// oclassCUتABase class(O@abstract class)
/////////////////////////////////////////////////////////////////////////////

CGameState::CGameState(CGame *g)
{
	game = g; 	// ]wgamepointer
}

void CGameState::GotoGameState(int state)
{
	game->SetGameState(state);
}

void CGameState::OnDraw() // Template Method
{
	Show();
}

void CGameState::OnCycle() // Template Method
{
	Move();
	Show();
}

/////////////////////////////////////////////////////////////////////////////
// CGame: Game Class
// oclassOCfacadeAOMFCPUӹCA١ApGW[δ
// CAܡAiHκ޳oclassP@C
/////////////////////////////////////////////////////////////////////////////

CGame game;

CGame::CGame()
: NUM_GAME_STATES(3)
{
	running = true;
	gameStateTable[GAME_STATE_INIT] = new CGameStateInit(this);
	gameStateTable[GAME_STATE_RUN]  = new CGameStateRun(this);
	gameStateTable[GAME_STATE_OVER] = new CGameStateOver(this);
	//gameState = gameStateTable[GAME_STATE_INIT];
	gameState = NULL;
}

CGame::~CGame()
{
	for (int i = 0; i < NUM_GAME_STATES; i++)
		delete gameStateTable[i];
}

bool CGame::IsRunning()
{
	return running;
}

void CGame::OnDraw()
{
	CDDraw::BltBackColor(DEFAULT_BG_COLOR);	// N Back Plain 
	gameState->OnDraw();					// ܹCCӤ
	if (!running) {
		//
		// pGbȰAAhCtrl-Q...
		//
		CMovingBitmap bmp;
		bmp.LoadBitmap(IDB_CONTINUE);
		bmp.SetTopLeft(0,0);
		bmp.ShowBitmap();
	}
	CDDraw::BltBackToPrimary();				// N Back Plain Kù
}

void  CGame::OnFilePause()
{
	if (running)
		CAudio::Instance()->Pause();
	else
		CAudio::Instance()->Resume();
	running = !running;
}

bool CGame::OnIdle()  // ק\ɳ̦nnקOnIdle()AקMove()Show()n
{
	//
	// CO_Ȱ
	//
	if (!running)
		return false;
	//
	// HUOCDj
	//
	CDDraw::BltBackColor(DEFAULT_BG_COLOR);	// N Back Plain 
	gameState->OnCycle();
	CDDraw::BltBackToPrimary();				// N Back Plain Kù
	//
	// HU{Ci檺tסA`NƶG
	// 1. Debug modeiH˵C@jᱼɶAOɶtC
	// 2. qW}OnIdle()ܦAɶw33msAiRAɶiCtC
	//
	if (SHOW_GAME_CYCLE_TIME)
		TRACE("Ellipse time for the %d th cycle=%d \n", CSpecialEffect::GetCurrentTimeCount(),CSpecialEffect::GetEllipseTime());
	CSpecialEffect::DelayFromSetCurrentTime(GAME_CYCLE_TIME);
	CSpecialEffect::SetCurrentTime();	// ]w}OnIdle()ɶ
	return true;
}

void CGame::OnInitialUpdate()	// OnInitialUpdate() ub{@}lɰ@
{
	//
	// Ұʶü
	//
	srand((unsigned)time(NULL));
	//
	// }DirectXøϤ
	//
	CDDraw::Init(SIZE_X, SIZE_Y);							// ]wCѪR
	//
	// }DirectXĤ
	//
	if (!CAudio::Instance()->Open()) {						// }ҭĤ
		AfxMessageBox("Audio Interface Failed (muted)");	// LĤ
		return;
	}
	//
	// IsCӪAOnInitialUpdate
	//
	for (int i = 0; i < NUM_GAME_STATES; i++)
		gameStateTable[i]->OnInit();
	//
	// Switch to the first state
	//
	SetGameState(GAME_STATE_INIT);
}

void CGame::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (running)
		if ((nFlags & 0x4000) == 0) // hauto repeat
			gameState->OnKeyDown(nChar, nRepCnt, nFlags);
#ifdef _UNITTEST					// invike unit test if _UNITTEST is defined
	void runTest();
	if (nChar == 'T')
		runTest();
#endif
}

void CGame::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (running)
		gameState->OnKeyUp(nChar, nRepCnt, nFlags);
}

void CGame::OnKillFocus()
{
	CAudio::Instance()->Pause();
	running = false;
}

void CGame::OnLButtonDown(UINT nFlags, CPoint point)
{
	if (running)
		gameState->OnLButtonDown(nFlags, point);
}

void CGame::OnRButtonDown(UINT nFlags, CPoint point)
{
	if (running)
		gameState->OnRButtonDown(nFlags, point);
}

void CGame::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (running)
		gameState->OnLButtonUp(nFlags, point);
}

void CGame::OnMouseMove(UINT nFlags, CPoint point)
{
	if (running)
		gameState->OnMouseMove(nFlags, point);
}

void CGame::OnRButtonUp(UINT nFlags, CPoint point)
{
	if (running)
		gameState->OnRButtonUp(nFlags, point);
}

void CGame::SetGameState(int state)
{
	ASSERT(state >=0 && state < NUM_GAME_STATES);
	gameState = gameStateTable[state];
	gameState->OnBeginState();
	OnDraw();
	CSpecialEffect::SetCurrentTime();
	running = true;
}

/////////////////////////////////////////////////////////////////////////////
// CSpecialEffect: Specail Effect functions
// @몺Cäݪާ@oӪA]iHL
/////////////////////////////////////////////////////////////////////////////

DWORD CSpecialEffect::ctime=0;
int   CSpecialEffect::ctimeCount=0;

void CSpecialEffect::Delay(DWORD ms)
{
	DWORD ctime = timeGetTime();
	int waitms;
	waitms = ms - (timeGetTime() - ctime);
	if (waitms > 0)
		Sleep(waitms);
}

void CSpecialEffect::DelayFromSetCurrentTime(DWORD ms)
{
	int waitms;
	waitms = ms - (timeGetTime() - ctime);
	if (waitms > 0)
		Sleep(waitms);
}

void CSpecialEffect::SetCurrentTime()
{
	ctime = timeGetTime();
	ctimeCount++;
}

DWORD CSpecialEffect::GetEllipseTime()
{
	return timeGetTime()-ctime;
}

int CSpecialEffect::GetCurrentTimeCount()
{
	return ctimeCount;
}

/////////////////////////////////////////////////////////////////////////////
// CDDraw: Direct Draw Object
// oclass|إDirectDrawAHѨLclassϥ
// oclass{OCøϤAiHL
/////////////////////////////////////////////////////////////////////////////

HDC							CDDraw::hdc;
CDC							CDDraw::cdc;
CView						*CDDraw::pCView;
LPDIRECTDRAW2				CDDraw::lpDD;
LPDIRECTDRAWCLIPPER			CDDraw::lpClipperPrimary;   
LPDIRECTDRAWCLIPPER			CDDraw::lpClipperBack;   
LPDIRECTDRAWSURFACE			CDDraw::lpDDSPrimary;
LPDIRECTDRAWSURFACE			CDDraw::lpDDSBack;
HRESULT						CDDraw::ddrval;
bool						CDDraw::fullscreen;
int							CDDraw::size_x, CDDraw::size_y;
CDDraw						CDDraw::ddraw;
vector<int>					CDDraw::BitmapID;
vector<CRect>				CDDraw::BitmapRect;
vector<COLORREF>			CDDraw::BitmapColorKey;
vector<LPDIRECTDRAWSURFACE>	CDDraw::lpDDS;

CDDraw::CDDraw()
{
	pCView=NULL;
	lpClipperPrimary=lpClipperBack=NULL;
	lpDDSPrimary=lpDDSBack=NULL;
	fullscreen = OPEN_AS_FULLSCREEN;
	lpDD=NULL;
}

CDDraw::~CDDraw()
{
	ReleaseSurface();
	if (lpDD)
		lpDD->Release();
	pCView=NULL; lpDD=NULL;
	lpClipperPrimary=lpClipperBack=NULL;
	lpDDSPrimary=lpDDSBack=NULL;
	TRACE("~CDDraw()\n");
}

void CDDraw::BltBackColor(DWORD color)
{
	if (lpDDSBack) {
		if (lpDDSBack->IsLost())
			RestoreSurface();
		DDBLTFX ddbltfx;
		ddbltfx.dwSize = sizeof(ddbltfx);
		ddbltfx.dwFillColor = MatchColorKey(lpDDSBack,color);
		ddrval = lpDDSBack->Blt(NULL,NULL,NULL,DDBLT_COLORFILL|DDBLT_WAIT,&ddbltfx);
		CheckDDFail("BltBackColor: blittng failed");
	}
}

void CDDraw::BltBackToPrimary()
{
	if (!pCView || !lpDDSPrimary || !lpDDSBack)
		return;
	if (fullscreen) {
		CRect ClientRect;
		GetClientRect(ClientRect);
		if (lpDDSBack->IsLost())
			RestoreSurface();
		if (lpDDSPrimary->IsLost())
			RestoreSurface();
		ddrval = lpDDSPrimary->Blt(ClientRect, lpDDSBack, ClientRect, DDBLTFAST_WAIT, NULL);
		CheckDDFail("Blt Back to primary failed");
	} else {
		if (lpDDSBack->IsLost())
			RestoreSurface();
		if (lpDDSPrimary->IsLost())
			RestoreSurface();
		CRect ClientRect;
		GetClientRect(ClientRect);
		CRect WindowRect;
		pCView->GetWindowRect(&WindowRect);
		WindowRect.right = WindowRect.left + size_x;
		WindowRect.bottom = WindowRect.top + size_y;
		ddrval = lpDDSPrimary->Blt(WindowRect, lpDDSBack, ClientRect, DDBLTFAST_WAIT, NULL);
		CheckDDFail("Blt Back to primary failed");
	}
}

void CDDraw::BltBitmapToBack(unsigned SurfaceID, int x, int y)
{
	GAME_ASSERT(lpDDSBack && (SurfaceID < lpDDS.size()) && lpDDS[SurfaceID], "Internal Error: Incorrect SurfaceID in BltBitmapToBack");
	CRect TargetRect;
	TargetRect.left = x;
	TargetRect.top = y;
	TargetRect.right = x + BitmapRect[SurfaceID].right-BitmapRect[SurfaceID].left;
	TargetRect.bottom = y + BitmapRect[SurfaceID].bottom-BitmapRect[SurfaceID].top;
	int blt_flag;
	if (BitmapColorKey[SurfaceID] != CLR_INVALID)
		blt_flag = DDBLT_WAIT | DDBLT_KEYSRC;
	else
		blt_flag = DDBLT_WAIT;
	if (lpDDSBack->IsLost())
		RestoreSurface();
	if (lpDDS[SurfaceID]->IsLost())
		RestoreSurface();
	ddrval = lpDDSBack->Blt(TargetRect, lpDDS[SurfaceID],NULL, blt_flag, NULL);
	CheckDDFail("Blt Bitmap to Back Failed");
}

void CDDraw::BltBitmapToBitmap(unsigned SourceID, unsigned TargetID, int x, int y)
{
	GAME_ASSERT((SourceID < lpDDS.size()) && (TargetID < lpDDS.size()) && (SourceID != TargetID), "Internal Error: Incorrect SourceID in BltBitmapToBitmap");
	int blt_flag;
	if (BitmapColorKey[SourceID] != CLR_INVALID)
		blt_flag = DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY;
	else
		blt_flag = DDBLTFAST_WAIT;
	if (lpDDS[SourceID]->IsLost())
		RestoreSurface();
	if (lpDDS[TargetID]->IsLost())
		RestoreSurface();
	ddrval = lpDDS[TargetID]->BltFast(x, y, lpDDS[SourceID], NULL, blt_flag );
	CheckDDFail("Blt Bitmap to Back Failed");
}

bool CDDraw::CreateSurface()
{
	//
	// Since all surfaces are created and loaded, this operation can be very slow.
	//
	SetCursor(AfxGetApp()->LoadStandardCursor(IDC_WAIT));
	ReleaseSurface();
	if (fullscreen) {
		if (!CreateSurfaceFullScreen())
			return false;
	} else {
		if (!CreateSurfaceWindowed())
			return false;
	}
	for (unsigned i = 0; i < lpDDS.size(); i++) {
		LoadBitmap(i, BitmapID[i]);
		SetColorKey(i, BitmapColorKey[i]);
	}
	return true;
}

bool CDDraw::CreateSurfaceFullScreen()
{
    ddrval = lpDD->SetCooperativeLevel(AfxGetMainWnd()->m_hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
    CheckDDFail("Can not SetCooperativeLevel Exclusive");
    ddrval = lpDD->SetDisplayMode(size_x, size_y, 16, 0, 0);
	if (ddrval != DD_OK) {
	    ddrval = lpDD->SetCooperativeLevel(AfxGetMainWnd()->m_hWnd, DDSCL_NORMAL);
		CheckDDFail("Can not SetCooperativeLevel Normal");
		return false;
	}
	CheckDDFail("SetDisplayMode FAILED");

	DDSURFACEDESC ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	CheckDDFail("Create Primary Surface failed");

	// Create clippers for the primary and back surfaces
    // ddrval = lpDD->CreateClipper(0, &lpClipperPrimary, NULL);
	// CheckDDFail("Create Primay Surface Clipper FAILED"); 
    ddrval = lpDD->CreateClipper(0, &lpClipperBack, NULL);
	CheckDDFail("Create Back Surface Clipper FAILED"); 

	typedef struct {
		RGNDATAHEADER hdr;
		RECT rgndata[4]; 
    } CLIPLIST, *LPCLIPLIST; 
	CLIPLIST    ClipList; 
	RECT rc;
	SetRect(&rc, 0, 0, size_x, size_y);  
    ClipList.hdr.dwSize = sizeof(RGNDATAHEADER); 
    ClipList.hdr.iType = RDH_RECTANGLES;
	ClipList.hdr.nCount = 1; 
    ClipList.hdr.nRgnSize = 0; 
    memcpy(&ClipList.hdr.rcBound, &rc, sizeof(RECT)); 
    memcpy(&ClipList.rgndata, &rc, sizeof(RECT)); 
	ddrval = lpClipperBack->SetClipList((LPRGNDATA)&ClipList,0); 
	CheckDDFail("SetHWnd FAILED"); 

	// Create Back (Secondary) Surface
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwHeight = size_y; ddsd.dwWidth = size_x;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSBack, NULL );
	CheckDDFail("Can not create back plain");

	// In fullscreen mode, the clipper for primary surafce is fixed to screen size
	ddrval = lpDDSPrimary->SetClipper(lpClipperBack);
	CheckDDFail("SetClipper FAILED");  
    ddrval = lpDDSBack->SetClipper(lpClipperBack);
	CheckDDFail("SetClipper FAILED");  
	lpClipperBack->Release();

	BltBackColor(RGB(0,0,0));
	BltBackToPrimary();
	return true;
}

bool CDDraw::CreateSurfaceWindowed()
{
    ddrval = lpDD->SetCooperativeLevel(AfxGetMainWnd()->m_hWnd, DDSCL_NORMAL);
    CheckDDFail("Can not SetCooperativeLevel ");

	DDSURFACEDESC ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	CheckDDFail("Create Primary Surface failed");

	// Create clippers for the primary and back surfaces
    ddrval = lpDD->CreateClipper(0, &lpClipperPrimary, NULL);
	CheckDDFail("Create Primay Surface Clipper FAILED"); 
    ddrval = lpDD->CreateClipper(0, &lpClipperBack, NULL);
	CheckDDFail("Create Back Surface Clipper FAILED"); 

    ddrval = lpClipperPrimary->SetHWnd(0, AfxGetMainWnd()->m_hWnd);
	CheckDDFail("Primary Surface SetHWnd FAILED"); 

	typedef struct {
		RGNDATAHEADER hdr;
		RECT rgndata[4]; 
    } CLIPLIST, *LPCLIPLIST; 
	CLIPLIST    ClipList; 
	RECT rc;
	SetRect(&rc, 0, 0, size_x, size_y);  
    ClipList.hdr.dwSize = sizeof(RGNDATAHEADER); 
    ClipList.hdr.iType = RDH_RECTANGLES;
	ClipList.hdr.nCount = 1; 
    ClipList.hdr.nRgnSize = 0; 
    memcpy(&ClipList.hdr.rcBound, &rc, sizeof(RECT)); 
    memcpy(&ClipList.rgndata, &rc, sizeof(RECT)); 
	ddrval = lpClipperBack->SetClipList((LPRGNDATA)&ClipList,0); 
	CheckDDFail("SetHWnd FAILED"); 

	// Create Back (Secondary) Surface
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwHeight = size_y; ddsd.dwWidth = size_x;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSBack, NULL );
	CheckDDFail("Can not create back plain");

	ddrval = lpDDSPrimary->SetClipper(lpClipperPrimary);
	CheckDDFail("SetClipper FAILED");  
	lpClipperPrimary->Release();
    ddrval = lpDDSBack->SetClipper(lpClipperBack);
	CheckDDFail("SetClipper FAILED");  
	lpClipperBack->Release();

	BltBackColor(RGB(0,0,0));
	return true;
}

void CDDraw::GetClientRect(CRect &r)
{
	r = CRect(0,0,size_x,size_y);
}

void CDDraw::Init(int sx, int sy)
{
	// set target screen size
	size_x = sx, size_y = sy;
	// init lpDD
	LPDIRECTDRAW lpDD0;
    ddrval = DirectDrawCreate(NULL, &lpDD0, NULL);
	CheckDDFail("DDraw create failed");
	ddrval = lpDD0->QueryInterface(IID_IDirectDraw2, (void **) &lpDD);
	CheckDDFail("DDraw surface 2 create failed");
	// init pCView
	POSITION pos = AfxGetApp()->GetFirstDocTemplatePosition();
	CDocTemplate *doc_t = AfxGetApp()->GetNextDocTemplate(pos);
	pos = doc_t->GetFirstDocPosition();
	CDocument *doc = doc_t->GetNextDoc(pos);
	pos = doc->GetFirstViewPosition();
	pCView = doc->GetNextView(pos);
	// init surfaces
	SetFullScreen(fullscreen);
}

void CDDraw::LoadBitmap(int i, int IDB_BITMAP)
{
	CDC mDC;
	CBitmap bitmap;
	BITMAP bitmapSize;
	bitmap.LoadBitmap(IDB_BITMAP);
	mDC.CreateCompatibleDC(NULL);
	CBitmap* pOldBitmap = mDC.SelectObject(&bitmap);
	bitmap.GetBitmap(&bitmapSize);
	DDSURFACEDESC ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize = sizeof( ddsd );
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	BitmapRect[i].bottom = ddsd.dwHeight = bitmapSize.bmHeight;
	BitmapRect[i].right = ddsd.dwWidth = bitmapSize.bmWidth;
	ddrval = lpDD->CreateSurface(&ddsd, &lpDDS[i], NULL);
	CheckDDFail("Create Bitmap Surface Failed");
	HDC hdc;
    ddrval= lpDDS[i]->GetDC(&hdc);
	CheckDDFail("Get surface HDC failed");
	CDC cdc;
	cdc.Attach(hdc);
	cdc.BitBlt(0,0,bitmapSize.bmWidth,bitmapSize.bmHeight, &mDC,0,0,SRCCOPY);
	cdc.Detach();
	lpDDS[i]->ReleaseDC(hdc);
	// avoid memory leak
	// According to spec, mDC should delete itself automatically.  However,
	// it appears that we have to do it explictly.
	mDC.SelectObject(&pOldBitmap);	
	mDC.DeleteDC();
	bitmap.DeleteObject();
}

DWORD CDDraw::MatchColorKey(LPDIRECTDRAWSURFACE lpDDSurface, COLORREF color)
{
    DDSURFACEDESC ddsd;
	HDC hdc;
	HRESULT hres;
	COLORREF rgbT= CLR_INVALID;
	DWORD dw=CLR_INVALID,mask=(DWORD)~0;
	if (lpDDSurface && color != CLR_INVALID) {
		if (lpDDSurface->GetDC(&hdc) == DD_OK) {
			rgbT = GetPixel(hdc, 0, 0);             // save (0,0) pixel value
			SetPixel(hdc, 0, 0, color);             // set our value
			lpDDSurface->ReleaseDC(hdc);
		}
		ddsd.dwSize = sizeof(ddsd);
		while ((hres = lpDDSurface->Lock(NULL, &ddsd, 0, NULL)) == DDERR_WASSTILLDRAWING);
		if (hres == DD_OK) {
			dw = *(DWORD *)ddsd.lpSurface;          // get (0,0) data
			if (ddsd.ddpfPixelFormat.dwRGBBitCount < 32)
				mask = (1 << ddsd.ddpfPixelFormat.dwRGBBitCount)-1;
			dw &= mask;								// mask it to bpp
			lpDDSurface->Unlock(NULL);
		}
		if (lpDDSurface->GetDC(&hdc) == DD_OK)	{
			SetPixel(hdc, 0, 0, rgbT);				// restore (0,0) pixel value
			lpDDSurface->ReleaseDC(hdc);
		}
	}
	return dw;
}

CDC* CDDraw::GetBackCDC()
{
	if (lpDDSBack->IsLost())
		RestoreSurface();
    ddrval= lpDDSBack->GetDC(&hdc);
	CheckDDFail("Get back surface HDC failed");
	cdc.Attach(hdc);
	return &cdc;
}

int CDDraw::RegisterBitmap(int IDB_BITMAP, COLORREF ColorKey)
{
	unsigned i;
	for (i = 0; i < lpDDS.size(); i++)
		if (BitmapID[i] == IDB_BITMAP)
			return i;
	//
	// Enlarge the size of vectors
	//
	BitmapID.push_back(IDB_BITMAP);
	BitmapColorKey.push_back(ColorKey);
	BitmapRect.push_back(CRect(0,0,0,0));
	lpDDS.push_back(NULL);
	LoadBitmap(i, IDB_BITMAP);
	SetColorKey(i, ColorKey);
	return i;
}
	
void CDDraw::ReleaseBackCDC()
{
	cdc.Detach();
	ddrval = lpDDSBack->ReleaseDC(hdc);
	CheckDDFail("Release back HDC failed");
}

void CDDraw::ReleaseSurface()
{
    if (lpDD)
    {
		for (unsigned i = 0; i < lpDDS.size(); i++)
			if (lpDDS[i]) {
				lpDDS[i]->Release();
				lpDDS[i] = NULL;
			}
        if (lpDDSBack)
        {
            lpDDSBack->Release();
            lpDDSBack = NULL;
        }
        if (lpDDSPrimary)
        {
            lpDDSPrimary->Release();
            lpDDSPrimary = NULL;
        }
	}
}

void CDDraw::RestoreSurface()
{
	//
	// Since all surfaces are restored and reloaded, this operation can be very slow.
	//
	SetCursor(AfxGetApp()->LoadStandardCursor(IDC_WAIT));
    if (lpDD != NULL)
    {
		CreateSurface();
		while (lpDDSBack->IsLost() || lpDDSPrimary->IsLost()) {
			Sleep(100);
			CreateSurface();
		}
    }
}

void CDDraw::SetColorKey(unsigned SurfaceID, COLORREF color)
{
	if (color != CLR_INVALID) {
		DDCOLORKEY ddck;
		ddck.dwColorSpaceLowValue  = MatchColorKey(lpDDS[SurfaceID], color);
		ddck.dwColorSpaceHighValue = ddck.dwColorSpaceLowValue;
		ddrval = lpDDS[SurfaceID]->SetColorKey(DDCKEY_SRCBLT, &ddck);
		CheckDDFail("Can not Set Color Key");
	}
}

bool CDDraw::SetFullScreen(bool isFullScreen)
{
	fullscreen = isFullScreen;
	return CreateSurface();
}

void CDDraw::CheckDDFail(char *s)
{
	if (ddrval != DD_OK) {
		TRACE("Error Code: %d (%s)\n", ddrval, s);
		//
		// For some unknown reason, in Win98, ddrval can be changed
		//	 after AfxMessageBox() is called, resulting an unknown
		//	 error message.
		//
		AfxMessageBox(s);
		static int ErrorCode[] = {
			DDERR_ALREADYINITIALIZED          ,DDERR_BLTFASTCANTCLIP             ,DDERR_CANNOTATTACHSURFACE         ,DDERR_CANNOTDETACHSURFACE         ,
			DDERR_CANTCREATEDC                ,DDERR_CANTDUPLICATE               ,DDERR_CANTLOCKSURFACE             ,DDERR_CANTPAGELOCK                ,
			DDERR_CANTPAGEUNLOCK              ,DDERR_CLIPPERISUSINGHWND          ,DDERR_COLORKEYNOTSET              ,DDERR_CURRENTLYNOTAVAIL           ,
			DDERR_DCALREADYCREATED            ,DDERR_DIRECTDRAWALREADYCREATED    ,DDERR_EXCEPTION                   ,DDERR_EXCLUSIVEMODEALREADYSET     ,
			DDERR_GENERIC                     ,DDERR_HEIGHTALIGN                 ,DDERR_HWNDALREADYSET              ,DDERR_HWNDSUBCLASSED              ,
			DDERR_IMPLICITLYCREATED           ,DDERR_INCOMPATIBLEPRIMARY         ,DDERR_INVALIDCAPS                 ,DDERR_INVALIDCLIPLIST             ,
			DDERR_INVALIDDIRECTDRAWGUID       ,DDERR_INVALIDMODE                 ,DDERR_INVALIDOBJECT               ,DDERR_INVALIDPARAMS               ,
			DDERR_INVALIDPIXELFORMAT          ,DDERR_INVALIDPOSITION             ,DDERR_INVALIDRECT                 ,DDERR_INVALIDSURFACETYPE          ,
			DDERR_LOCKEDSURFACES              ,DDERR_NO3D                        ,DDERR_NOALPHAHW                   ,DDERR_NOBLTHW                     ,
			DDERR_NOCLIPLIST                  ,DDERR_NOCLIPPERATTACHED           ,DDERR_NOCOLORCONVHW               ,DDERR_NOCOLORKEY                  ,
			DDERR_NOCOLORKEYHW                ,DDERR_NOCOOPERATIVELEVELSET       ,DDERR_NODC                        ,DDERR_NODDROPSHW                  ,
			DDERR_NODIRECTDRAWHW              ,DDERR_NODIRECTDRAWSUPPORT         ,DDERR_NOEMULATION                 ,DDERR_NOEXCLUSIVEMODE             ,
			DDERR_NOFLIPHW                    ,DDERR_NOGDI                       ,DDERR_NOHWND                      ,DDERR_NOMIPMAPHW                  ,
			DDERR_NOMIRRORHW                  ,DDERR_NOOVERLAYDEST               ,DDERR_NOOVERLAYHW                 ,DDERR_NOPALETTEATTACHED           ,
			DDERR_NOPALETTEHW                 ,DDERR_NORASTEROPHW                ,DDERR_NOROTATIONHW                ,DDERR_NOSTRETCHHW                 ,
			DDERR_NOT4BITCOLOR                ,DDERR_NOT4BITCOLORINDEX           ,DDERR_NOT8BITCOLOR                ,DDERR_NOTAOVERLAYSURFACE          ,
			DDERR_NOTEXTUREHW                 ,DDERR_NOTFLIPPABLE                ,DDERR_NOTFOUND                    ,DDERR_NOTINITIALIZED              ,
			DDERR_NOTLOCKED                   ,DDERR_NOTPAGELOCKED               ,DDERR_NOTPALETTIZED               ,DDERR_NOVSYNCHW                   ,
			DDERR_NOZBUFFERHW                 ,DDERR_NOZOVERLAYHW                ,DDERR_OUTOFCAPS                   ,DDERR_OUTOFMEMORY                 ,
			DDERR_OUTOFVIDEOMEMORY            ,DDERR_OVERLAYCANTCLIP             ,DDERR_OVERLAYCOLORKEYONLYONEACTIVE,DDERR_OVERLAYNOTVISIBLE           ,
			DDERR_PALETTEBUSY                 ,DDERR_PRIMARYSURFACEALREADYEXISTS ,DDERR_REGIONTOOSMALL              ,DDERR_SURFACEALREADYATTACHED      ,
			DDERR_SURFACEALREADYDEPENDENT     ,DDERR_SURFACEBUSY                 ,DDERR_SURFACEISOBSCURED           ,DDERR_SURFACELOST                 ,
			DDERR_SURFACENOTATTACHED          ,DDERR_TOOBIGHEIGHT                ,DDERR_TOOBIGSIZE                  ,DDERR_TOOBIGWIDTH                 ,
			DDERR_UNSUPPORTED                 ,DDERR_UNSUPPORTEDFORMAT           ,DDERR_UNSUPPORTEDMASK             ,DDERR_UNSUPPORTEDMODE             ,
			DDERR_VERTICALBLANKINPROGRESS     ,DDERR_WASSTILLDRAWING             ,DDERR_WRONGMODE                   ,DDERR_XALIGN
		};
		static char *ErrorMsg[] = {
			"DDERR_ALREADYINITIALIZED          ","DDERR_BLTFASTCANTCLIP             ","DDERR_CANNOTATTACHSURFACE         ","DDERR_CANNOTDETACHSURFACE         ",
			"DDERR_CANTCREATEDC                ","DDERR_CANTDUPLICATE               ","DDERR_CANTLOCKSURFACE             ","DDERR_CANTPAGELOCK                ",
			"DDERR_CANTPAGEUNLOCK              ","DDERR_CLIPPERISUSINGHWND          ","DDERR_COLORKEYNOTSET              ","DDERR_CURRENTLYNOTAVAIL           ",
			"DDERR_DCALREADYCREATED            ","DDERR_DIRECTDRAWALREADYCREATED    ","DDERR_EXCEPTION                   ","DDERR_EXCLUSIVEMODEALREADYSET     ",
			"DDERR_GENERIC                     ","DDERR_HEIGHTALIGN                 ","DDERR_HWNDALREADYSET              ","DDERR_HWNDSUBCLASSED              ",
			"DDERR_IMPLICITLYCREATED           ","DDERR_INCOMPATIBLEPRIMARY         ","DDERR_INVALIDCAPS                 ","DDERR_INVALIDCLIPLIST             ",
			"DDERR_INVALIDDIRECTDRAWGUID       ","DDERR_INVALIDMODE                 ","DDERR_INVALIDOBJECT               ","DDERR_INVALIDPARAMS               ",
			"DDERR_INVALIDPIXELFORMAT          ","DDERR_INVALIDPOSITION             ","DDERR_INVALIDRECT                 ","DDERR_INVALIDSURFACETYPE          ",
			"DDERR_LOCKEDSURFACES              ","DDERR_NO3D                        ","DDERR_NOALPHAHW                   ","DDERR_NOBLTHW                     ",
			"DDERR_NOCLIPLIST                  ","DDERR_NOCLIPPERATTACHED           ","DDERR_NOCOLORCONVHW               ","DDERR_NOCOLORKEY                  ",
			"DDERR_NOCOLORKEYHW                ","DDERR_NOCOOPERATIVELEVELSET       ","DDERR_NODC                        ","DDERR_NODDROPSHW                  ",
			"DDERR_NODIRECTDRAWHW              ","DDERR_NODIRECTDRAWSUPPORT         ","DDERR_NOEMULATION                 ","DDERR_NOEXCLUSIVEMODE             ",
			"DDERR_NOFLIPHW                    ","DDERR_NOGDI                       ","DDERR_NOHWND                      ","DDERR_NOMIPMAPHW                  ",
			"DDERR_NOMIRRORHW                  ","DDERR_NOOVERLAYDEST               ","DDERR_NOOVERLAYHW                 ","DDERR_NOPALETTEATTACHED           ",
			"DDERR_NOPALETTEHW                 ","DDERR_NORASTEROPHW                ","DDERR_NOROTATIONHW                ","DDERR_NOSTRETCHHW                 ",
			"DDERR_NOT4BITCOLOR                ","DDERR_NOT4BITCOLORINDEX           ","DDERR_NOT8BITCOLOR                ","DDERR_NOTAOVERLAYSURFACE          ",
			"DDERR_NOTEXTUREHW                 ","DDERR_NOTFLIPPABLE                ","DDERR_NOTFOUND                    ","DDERR_NOTINITIALIZED              ",
			"DDERR_NOTLOCKED                   ","DDERR_NOTPAGELOCKED               ","DDERR_NOTPALETTIZED               ","DDERR_NOVSYNCHW                   ",
			"DDERR_NOZBUFFERHW                 ","DDERR_NOZOVERLAYHW                ","DDERR_OUTOFCAPS                   ","DDERR_OUTOFMEMORY                 ",
			"DDERR_OUTOFVIDEOMEMORY            ","DDERR_OVERLAYCANTCLIP             ","DDERR_OVERLAYCOLORKEYONLYONEACTIVE","DDERR_OVERLAYNOTVISIBLE           ",
			"DDERR_PALETTEBUSY                 ","DDERR_PRIMARYSURFACEALREADYEXISTS ","DDERR_REGIONTOOSMALL              ","DDERR_SURFACEALREADYATTACHED      ",
			"DDERR_SURFACEALREADYDEPENDENT     ","DDERR_SURFACEBUSY                 ","DDERR_SURFACEISOBSCURED           ","DDERR_SURFACELOST                 ",
			"DDERR_SURFACENOTATTACHED          ","DDERR_TOOBIGHEIGHT                ","DDERR_TOOBIGSIZE                  ","DDERR_TOOBIGWIDTH                 ",
			"DDERR_UNSUPPORTED                 ","DDERR_UNSUPPORTEDFORMAT           ","DDERR_UNSUPPORTEDMASK             ","DDERR_UNSUPPORTEDMODE             ",
			"DDERR_VERTICALBLANKINPROGRESS     ","DDERR_WASSTILLDRAWING             ","DDERR_WRONGMODE                   ","DDERR_XALIGN                      "
		};
		for (int i = 0; i < sizeof(ErrorCode)/sizeof(int); i++)
			if (ddrval == ErrorCode[i])
				GAME_ASSERT(0, ErrorMsg[i]);
		GAME_ASSERT(0, "Direct Draw Failed due to unknown error code !!!");
	}
}

/////////////////////////////////////////////////////////////////////////////
// CAudio
// oclassѭ(WaveMIDI)
// noIs(B)UدOAOiHUC{ON
// Note: The following memebers of CAudio are designed to be graceful so
//    that the game does not break down, when NO audio device is ready.
/////////////////////////////////////////////////////////////////////////////

CAudio CAudio::audio;

CAudio::CAudio()
: g_pSegment(20),g_pSegmentState(20),g_ResumePoint(20),
  g_isPaused(20,false),isMIDI(20)
{
	g_pLoader = NULL;
	g_pPerformance = NULL;
	//
	// The following initialization is may be unnecessary.
	//     The contents of the vector appear to be NULL already.
	//
	for (unsigned i=0; i < g_pSegment.size(); i++)
		g_pSegment[i] = NULL;
}

CAudio::~CAudio()
{
	Close();
	TRACE("~CAudio()\n");
}

void CAudio::Close()
{
	if (g_pPerformance) {
		g_pPerformance->Stop(
			NULL,   // Stop all segments.
			NULL,   // Stop all segment states.
			0,      // Do it immediately.
			0       // Flags.
		);
		g_pPerformance->CloseDown();
	}
	if (g_pLoader)
		g_pLoader->Release(); 
	if (g_pPerformance)
		g_pPerformance->Release();
	for (unsigned i=0; i < g_pSegment.size(); i++) {
		if (g_pSegmentState[i])
			g_pSegmentState[i]->Release();
		g_pSegmentState[i] = NULL;
		if (g_pSegment[i])
			g_pSegment[i]->Release();
		g_pSegment[i] = NULL;
	}
	g_pPerformance = NULL;
	g_pLoader = NULL;
    CoUninitialize();
}

CAudio *CAudio::Instance()
{
	return &audio;
}

bool CAudio::Load(unsigned id, char *lpzFileName)
{
	if (!g_pPerformance || !g_pLoader)
		return false;
	if (id >= g_pSegment.size()) {
		//
		// resize vector to fit
		//
		unsigned newsize = id * 2;
		g_pSegment.resize(newsize, NULL);
		g_pSegmentState.resize(newsize);
		g_ResumePoint.resize(newsize);
		g_isPaused.resize(newsize,false);
		isMIDI.resize(newsize);
	}
	if (g_pSegment[id])
		return false; // used id can not be used again
    WCHAR wszFile[_MAX_PATH];
	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, lpzFileName, -1, wszFile, _MAX_PATH);
    if (FAILED(g_pLoader->LoadObjectFromFile(
        CLSID_DirectMusicSegment,   // Class identifier.
        IID_IDirectMusicSegment8,   // ID of desired interface.
        wszFile,		            // Filename.
        (LPVOID*) &g_pSegment[id]   // Pointer that receives interface.
    )))
    {
		char s[200];
		sprintf(s, "Audio File %s NOT found!", lpzFileName);
		AfxMessageBox(s);
		g_pSegment[id] = NULL;
        return false;
    }
	g_pSegment[id]->Download(g_pPerformance);
	isMIDI[id] = (strcmp(&lpzFileName[strlen(lpzFileName)-4],".mid") == 0);
	return true;
}

bool CAudio::Open()
{
	if (!AUDIO_ENABLE)
		return false;
   	if (CoInitialize(NULL) != S_OK)
		return false;
    CoCreateInstance(CLSID_DirectMusicLoader, NULL, 
                     CLSCTX_INPROC, IID_IDirectMusicLoader8,
                     (void**)&g_pLoader);
	if (!g_pLoader)
		return false;
    CoCreateInstance(CLSID_DirectMusicPerformance, NULL,
                     CLSCTX_INPROC, IID_IDirectMusicPerformance8,
                     (void**)&g_pPerformance);
	if (!g_pPerformance)
		return false;
    g_pPerformance->InitAudio( 
        NULL,								// IDirectMusic interface not needed.
        NULL,								// IDirectSound interface not needed.
        AfxGetMainWnd()->m_hWnd,            // Window handle.
        DMUS_APATH_SHARED_STEREOPLUSREVERB, // Default audiopath type.
        64,									// Number of performance channels.
        DMUS_AUDIOF_ALL,					// Features on synthesizer.
        NULL								// Audio parameters; use defaults.
    );
	return true;
}

MUSIC_TIME CAudio::GetTimeOffset(const MUSIC_TIME mtNow,           // From GetTime
                         const MUSIC_TIME mtStartTime,     // From GetStartTime
                         const MUSIC_TIME mtStartPoint,    // From GetStartPoint
                         const MUSIC_TIME mtLoopStart,     // From GetLoopPoints
                         const MUSIC_TIME mtLoopEnd,       // From GetLoopPoints
                         const MUSIC_TIME mtLength,        // From GetLength
                         const DWORD dwLoopRepeats)        // From GetRepeats
{
    // Convert mtNow from absolute time to an offset
    // from when the segment started playing.
    LONGLONG llOffset = mtNow - (mtStartTime - mtStartPoint);
    // If mtLoopEnd is not zero, set llLoopEnd to mtLoopEnd;
    // otherwise use the segment length.
    LONGLONG llLoopEnd = mtLoopEnd ? mtLoopEnd : mtLength;
    LONGLONG llLoopStart = mtLoopStart;
    // Adjust offset to take looping into account.
    if ((dwLoopRepeats != 0) &&  (llLoopStart < llLoopEnd) &&  (llLoopEnd > mtStartPoint))
    {
        if ((dwLoopRepeats != DMUS_SEG_REPEAT_INFINITE)
          &&  (llOffset > (llLoopStart + (llLoopEnd - llLoopStart) *(signed)dwLoopRepeats)))
        {
            llOffset -= (llLoopEnd - llLoopStart) * dwLoopRepeats;
        }
        else if (llOffset > llLoopStart)
        {
            llOffset = llLoopStart + (llOffset - llLoopStart) % (llLoopEnd - llLoopStart);
        }
    }

    llOffset = min(llOffset, LONG_MAX);  // LONG_MAX is defined in Limits.h.
    return long(llOffset);
}

void CAudio::Pause()
{
	if (!g_pPerformance || !g_pLoader)
		return;
	for (unsigned id=0; id < g_pSegment.size(); id++) {
		if (g_pSegment[id] && (g_pPerformance->IsPlaying(g_pSegment[id], NULL) == S_OK)) {
			MUSIC_TIME mtNow,mtStartTime,mtStartPoint;
			MUSIC_TIME mtLoopStart,mtLoopEnd,mtLength;
			DWORD dwLoopRepeats;
			IDirectMusicSegment8 *pSegment = g_pSegment[id];
			g_pPerformance->GetTime(NULL, &mtNow);
			g_pSegmentState[id]->GetStartTime(&mtStartTime);
			pSegment->GetStartPoint(&mtStartPoint);
			pSegment->GetLoopPoints(&mtLoopStart, &mtLoopEnd);
			pSegment->GetLength(&mtLength);
			pSegment->GetRepeats(&dwLoopRepeats);
			g_ResumePoint[id]=GetTimeOffset(mtNow,mtStartTime,mtStartPoint,mtLoopStart,mtLoopEnd,mtLength,dwLoopRepeats);
			g_isPaused[id] = true;
			Stop(id);
		}
	}
}

void CAudio::Play(unsigned id, bool repeat)
{
	//
	// MIDI files are played as PRIMARY SEGMENTS.
	// WAVE files are played as SECONDARY SEGMENTS.
	//
	if (id >= g_pSegment.size() || !g_pPerformance || !g_pLoader || !g_pSegment[id])
		return;
	g_pSegment[id] ->SetRepeats(repeat ? DMUS_SEG_REPEAT_INFINITE : 0);
    g_pPerformance->PlaySegment(
        g_pSegment[id],		// Segment to play.
        (isMIDI[id]) ? DMUS_SEGF_DEFAULT : DMUS_SEGF_SECONDARY ,           // Flags.
        0,					// Start time; 0 is immediate.
        //NULL				// Pointer that receives segment state.
		&g_pSegmentState[id]
    );      
}

void CAudio::Resume()
{
	if (!g_pPerformance || !g_pLoader)
		return;
	for (unsigned id=0; id < g_pSegment.size(); id++) {
		if (g_pSegment[id] && g_isPaused[id]) {
			g_pSegment[id]->SetStartPoint(g_ResumePoint[id]);
			g_pPerformance->PlaySegment(
				g_pSegment[id],		// Segment to play.
				(isMIDI[id]) ? DMUS_SEGF_DEFAULT : DMUS_SEGF_SECONDARY,           // Flags.
				0,					// Start time; 0 is immediate.
				&g_pSegmentState[id]// Pointer that receives segment state.
			);      
			g_isPaused[id] = false;
		}
	}
}

void CAudio::Stop(unsigned id)
{
	if (id < g_pSegment.size() && g_pPerformance && g_pSegment[id])
		g_pPerformance->Stop(g_pSegment[id], NULL, 0, 0);
}
