/*
 * gamelib.h: ɮxCclassinterface
 * Copyright (C) 2002-2005 Woei-Kae Chen <wkc@csie.ntut.edu.tw>
 *
 * This file is part of game, a free game development framework for windows.
 *
 * game is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * game is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 *   2002-03-04 V3.1
 *		1. Add AUDIO_ENABLE flag.
 *   2004-03-02 V4.0
 *		1. Add CInteger class.
 *      2. Add CGameState abstract class and adjust CGame to apply
 *         state pattern for switching game states.
 *   2004-03-08 V4.1
 *      1. Add OPEN_AS_FULLSCREEN flag.
 *      2. The Game Engine now becomes a framework.
 *   2005-07-28
 *      1. Add GAME_ASSERT macro to display run time errors gracefully.
 *      2. Remove CDDarw::CheckSystem(). It is useless.
 *   2005-09-08
 *      1. Eliminate CSpecialEffect::Abort. It is useless now.
 *      2. Add SHOW_GAME_CYCLE_TIME flag.
 *      3. Add DEFAULT_BG_COLOR flag;
 *   2005-09-20 V4.2Beta1.
 *   2005-09-29 V4.2Beta2.
 *      1. Add MOUSEMOVE Handler for CGame and CGameState.
 *      2. Add _TRACE preprocessor flag for VC++.net.
 *   2006-02-08 V4.2
 *      1. Rename OnInitialUpdate() -> OnInit().
 *      2. Remove constantness of CAnimation::DELAY_COUNT.
 *      3. Enhance CAnimation to support SetDelayCount(), Reset(), and IsFinalBitmap().
 *      4. Remove CAnimation::GetLocation() and CMovingBitmap::GetLocation().
 *      5. Bitmap coordinate can no longer be set by CMovingBitmap::LoadBitmap().
 *         Defauts to (0,0).
*/

/////////////////////////////////////////////////////////////////////////////
// wqCi]wһP
/////////////////////////////////////////////////////////////////////////////

#define SIZE_X				 640		// ]wCeѪR׬640x480
#define SIZE_Y				 480		// `GYϥμзǪѪRסAhù
#define OPEN_AS_FULLSCREEN	 false		// O_Hù覡}ҹC
#define DEFAULT_BG_COLOR	 RGB(0,0,0)	// Cew]IC(¦)
#define AUDIO_ENABLE		 true		// ҰʭĤ
#define GAME_CYCLE_TIME		 33		    // C33ms]@MoveShow(C30)
#define SHOW_GAME_CYCLE_TIME false		// O_bdebug modecycle time

/////////////////////////////////////////////////////////////////////////////
// wqCGameCGameStateҨϥΪTӪA`
/////////////////////////////////////////////////////////////////////////////

enum GAME_STATES {
	GAME_STATE_INIT,
	GAME_STATE_RUN,
	GAME_STATE_OVER
};

/////////////////////////////////////////////////////////////////////////////
// Header for STL (Standard Template Library)
/////////////////////////////////////////////////////////////////////////////

#include <list>
#include <vector>
using namespace std;

/////////////////////////////////////////////////////////////////////////////
// ^{~Ϊmacro
// ƵGǫϥmacroHKOdɮצW٤Φ渹AQdebugC
/////////////////////////////////////////////////////////////////////////////

#define GAME_ASSERT(boolexp,str)											\
		if (!(boolexp)) {													\
			int id;															\
			char s[300]="";													\
			sprintf(s,"Game fatal error:\n\n%s\n\nFile: %s\n\nLine: %d"		\
				"\n\n(Press Retry to debug the application, "				\
				"if it is executed in debug mode.)"							\
				"\n(Press Cancel otherwise.)",								\
				 str , __FILE__,__LINE__);									\
			id = AfxMessageBox(s, MB_RETRYCANCEL);							\
			if (id == IDCANCEL)												\
				exit(1);													\
			AfxDebugBreak();												\
		}

/////////////////////////////////////////////////////////////////////////////
// oclassѮɶB~
// @몺Cäݪާ@oӪA]iH޳oclassϥΤk
/////////////////////////////////////////////////////////////////////////////

class CSpecialEffect {
public:
	static void  SetCurrentTime();					// xsثeɶctime
	static DWORD GetEllipseTime();					// Ūثeɶ - ctime
	static int   GetCurrentTimeCount();				// Ūxsctime
	static void  Delay(DWORD ms);					//  x ms
	static void  DelayFromSetCurrentTime(DWORD ms);	// ctime_A x ms
private:
	static DWORD ctime;
	static int	 ctimeCount;
};

/////////////////////////////////////////////////////////////////////////////
// oclass|إDirectDrawAHѨLclassϥ
// @몺Cäݪާ@oӪA]iH޳oclassϥΤk
/////////////////////////////////////////////////////////////////////////////

class CDDraw {
	friend class CMovingBitmap;
public:
	~CDDraw();
	static void  BltBackColor(DWORD);		// NBack plainۤWwC
	static void  BltBackToPrimary();		// NBack plainKPrimary plain
	static CDC*  GetBackCDC();				// oBack PlainDC (device context)
	static void  GetClientRect(CRect &r);	// o]wѪR
	static void  Init(int, int);			// Initialize direct draw
	static void  ReleaseBackCDC();			// Back PlainDC (device context)
	static bool  SetFullScreen(bool);		// ]wùҦ/Ҧ
private:
	CDDraw();								// private constructor
	static void  BltBitmapToBack(unsigned SurfaceID, int x, int y);
	static void  BltBitmapToBitmap(unsigned SourceID, unsigned TargetID, int x, int y);
	static void	 CheckDDFail(char *s);
	static bool  CreateSurface();
	static bool  CreateSurfaceFullScreen();
	static bool  CreateSurfaceWindowed();
	static void  LoadBitmap(int i, int IDB_BITMAP);
	static DWORD MatchColorKey(LPDIRECTDRAWSURFACE lpDDSurface, COLORREF color);
	static int   RegisterBitmap(int IDB_BITMAP, COLORREF ColorKey);
	static void  ReleaseSurface();
	static void  RestoreSurface();
	static void  SetColorKey(unsigned SurfaceID, COLORREF color);
    static HDC					hdc;
	static CDC					cdc;
	static CView				*pCView;
    static LPDIRECTDRAW2		lpDD;
	static LPDIRECTDRAWCLIPPER	lpClipperPrimary;   
	static LPDIRECTDRAWCLIPPER	lpClipperBack;   
	static LPDIRECTDRAWSURFACE	lpDDSPrimary;
	static LPDIRECTDRAWSURFACE	lpDDSBack;
	static vector<LPDIRECTDRAWSURFACE>	lpDDS;
    static HRESULT				ddrval;
	static vector<int>			BitmapID;
	static vector<CRect>		BitmapRect;
	static vector<COLORREF>		BitmapColorKey;
	static bool					fullscreen;
	static CDDraw				ddraw;
	static int					size_x, size_y;
};

/////////////////////////////////////////////////////////////////////////////
// oclassѰʺA(iH)ϧ
// CPublic InterfaceΪknAImplementationiH
/////////////////////////////////////////////////////////////////////////////

class CMovingBitmap {
public:
	CMovingBitmap();
	int   Height();						// oϧΪ
	int   Left();						// oϧΪW x y
	void  LoadBitmap(int,COLORREF=CLR_INVALID);
										// ]wϪȡGϪsByСBγz
	void  SetTopLeft(int,int);			// NϪWyв (x,y)
	void  ShowBitmap();					// N϶Kù
	void  ShowBitmap(CMovingBitmap &);	// N϶Kt@iϤW (ȨѯSγ~)
	int   Top();						// oϧΪW y y
	int   Width();						// oϧΪe
protected:
	CRect    location;			// location of the bitmap
	bool     isBitmapLoaded;	// whether a bitmap has been loaded
	unsigned SurfaceID;			// the surface id of this bitmap
};

/////////////////////////////////////////////////////////////////////////////
// oclass|إDirectX AudioAHѭ
// CPublic InterfaceΪknAImplementationiH
// `NƶGɮץiHWAVE(.wav)MIDI(.mid)AƭӭɮץiH
//           VX(P)AO䤤uVX@MIDIɡAMIDIɮפ
//           PɼApGPɼMIDIɪܡAe@MIDIɷ|Q
//           C
/////////////////////////////////////////////////////////////////////////////

class CAudio {
public:
	~CAudio();
	void           Close();						// Direct Sound
	static CAudio *Instance();					// oCAudioInstance
	bool           Load(unsigned, char *);		// JsinAwMIDIɮ
	bool           Open();						// }Direct Sound
	void		   Pause();						// ȰҦ
	void           Play(unsigned, bool=false);	// }lsinAwO_м
	void		   Resume();					// _Ȱ񪺭
	void           Stop(unsigned);				// sin
private:
	CAudio();									// private constructor
	static CAudio					   audio;	// ۰destruct
	MUSIC_TIME		 				   GetTimeOffset(MUSIC_TIME,MUSIC_TIME,MUSIC_TIME,MUSIC_TIME,MUSIC_TIME,MUSIC_TIME,DWORD);
	IDirectMusicLoader8				   *g_pLoader;
	IDirectMusicPerformance8		   *g_pPerformance;
	vector<IDirectMusicSegment8 *>	   g_pSegment;
	vector<IDirectMusicSegmentState *> g_pSegmentState;
	vector<long>					   g_ResumePoint;
	vector<bool>					   g_isPaused;
	vector<bool>					   isMIDI;
};

/////////////////////////////////////////////////////////////////////////////
// oclassѥiHʪʵe
// CPublic InterfaceΪknAImplementationiH
/////////////////////////////////////////////////////////////////////////////

class CAnimation {
public:
	CAnimation(int=10);				// Constructor (w]ʵeWvC1/3@i)
	void  AddBitmap(int,COLORREF=CLR_INVALID);
									// W[@iϧΦܰʵe(Ϫsγz)
	int   GetCurrentBitmapNumber();	// obbitmapOĴXbitmap
	int   Height();					// oʵe
	bool  IsFinalBitmap();			// ^ǥbbitmapO_̫@bitmap
	int   Left();					// oʵeW x y
	void  Move();					// Wvbitmap
	void  Reset();					// ]񶶧Ǧ^Ĥ@iϧ
	void  SetDelayCount(int);		// ]wʵetת`(VjVC)
	void  SetTopLeft(int,int);		// NʵeWyв (x,y)
	void  Show();					// NʵeKù
	int   Top();					// oʵeW y y
	int   Width();					// oʵee
private:
	list<CMovingBitmap>				bmp;			// list of CMovingBitmap
	list<CMovingBitmap>::iterator	bmp_iter;		// list iterator
	int								bmp_counter;	// xsbmp_iternbmp
	int								delay_counter;	// wʵetתpƾ
	int								delay_count;	// ʵetת`
	int								x, y;			// ʵey
};

/////////////////////////////////////////////////////////////////////////////
// oclassܾƹϧΪO
// CPublic InterfaceΪknAImplementationiH
/////////////////////////////////////////////////////////////////////////////

class CInteger {
public:
	CInteger(int=5);			// default 5 digits
	void Add(int n);			// W[ƭ
	int  GetInteger();			// ^Ǿƭ
	void LoadBitmap();			// J0..9έtϧ
	void SetInteger(int);		// ]wƭ
	void SetTopLeft(int,int);	// NʵeWyв (x,y)
	void ShowBitmap();			// NʵeKù
private:
	const int NUMDIGITS;			// @NUMDIGITSӦ
	static CMovingBitmap digit[11]; // xs0..9έtϧ(bitmap)
	int x, y;						// ܪy
	int n;							// ƭ
	bool isBmpLoaded;				// O_wgJϧ
};

/////////////////////////////////////////////////////////////////////////////
// ŧi|wqclass
/////////////////////////////////////////////////////////////////////////////

class CGame;
class CGameStateInit;
class CGameStateRun;
class CGameStateOver;

/////////////////////////////////////////////////////////////////////////////
// oclassCUتABase class(O@abstract class)
// CPublic InterfaceΪknAImplementationiH
/////////////////////////////////////////////////////////////////////////////

class CGameState {
public:
	CGameState(CGame *g);
	void OnDraw();			// Template Method
	void OnCycle();			// Templaye Method
	//
	// virtual functions, ~Ӫ̴implementation
	//
	virtual ~CGameState() {}								// virtual destructor
	virtual void OnBeginState() {}							// ]wCiJoӪAɩһݪ
	virtual void OnInit() {}								// AȤιϧγ]w
	virtual void OnKeyDown(UINT, UINT, UINT) {}				// BzLDownʧ@
	virtual void OnKeyUp(UINT, UINT, UINT) {}				// BzLUpʧ@
	virtual void OnLButtonDown(UINT nFlags, CPoint point) {}// Bzƹʧ@
	virtual void OnLButtonUp(UINT nFlags, CPoint point) {}	// Bzƹʧ@
	virtual void OnMouseMove(UINT nFlags, CPoint point) {}  // Bzƹʧ@ 
	virtual void OnRButtonDown(UINT nFlags, CPoint point) {}// Bzƹʧ@
	virtual void OnRButtonUp(UINT nFlags, CPoint point) {}	// Bzƹʧ@
protected:
	void GotoGameState(int);
	//
	// virtual functions, ~Ӫ̴implementation
	//
	virtual void Move() {}									// ʳoӪAC
	virtual void Show() = 0;								// ܳoӪACe
	CGame *game;
};

/////////////////////////////////////////////////////////////////////////////
// oclassOC֤ߡACi
// @몺Cäݪާ@oӪA]iH޳oclassϥΤk
/////////////////////////////////////////////////////////////////////////////

class CGame {
public:
	CGame();										// Constructor
	~CGame();										// Destructor
	bool IsRunning();								// ŪCO_bi椤
	void OnDraw();									// CGameViewOnDraw()
	void OnFilePause();								// CȰ
	void OnInitialUpdate();							// CȤιϧγ]w
	bool OnIdle();									// CDj
	void OnKeyDown(UINT, UINT, UINT);				// BzLDownʧ@
	void OnKeyUp(UINT, UINT, UINT);					// BzLUpʧ@
	void OnKillFocus();								// CQȰ
	void OnLButtonDown(UINT nFlags, CPoint point);	// Bzƹʧ@
	void OnLButtonUp(UINT nFlags, CPoint point);	// Bzƹʧ@
	void OnMouseMove(UINT nFlags, CPoint point);    // Bzƹʧ@ 
	void OnRButtonDown(UINT nFlags, CPoint point);	// Bzƹʧ@
	void OnRButtonUp(UINT nFlags, CPoint point);	// Bzƹʧ@
	void SetGameState(int);
private:
	bool			running;			// CO_bi椤(QPause)
	const int		NUM_GAME_STATES;	// CA(3ӪA)
	CGameState		*gameState;			// pointerVثeCA
	CGameState		*gameStateTable[3];	// CApointer
};
