/*
 * mygame.cpp: ɮxCclassimplementation
 * Copyright (C) 2002-2005 Woei-Kae Chen <wkc@csie.ntut.edu.tw>
 *
 * This file is part of game, a free game development framework for windows.
 *
 * game is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * game is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 *   2002-03-04 V3.1
 *          Add codes to demostrate the use of CMovingBitmap::ShowBitmap(CMovingBitmap &).
 *	 2004-03-02 V4.0
 *      1. Add CGameStateInit, CGameStateRun, and CGameStateOver to
 *         demonstrate the use of states.
 *      2. Demo the use of CInteger in CGameStateRun.
 *   2005-09-13
 *      Rewrite the codes for CBall and CEraser.
 *   2005-09-20 V4.2Beta1.
 *   2005-09-29 V4.2Beta2.
 *      1. Add codes to display IDC_GAMECURSOR in GameStateRun.
 *   2006-02-08 V4.2
 *      1. Revise sample screens to display in English only.
 *      2. Add code in CGameStateInit to demo the use of PostQuitMessage().
 *      3. Rename OnInitialUpdate() -> OnInit().
 *      4. Fix the bug that OnBeginState() of GameStateInit is not called.
 *      5. Replace AUDIO_CANYON as AUDIO_NTUT.
 *      6. Add help bitmap to CGameStateRun.
 */

#include "stdafx.h"
#include "Resource.h"
#include <mmsystem.h>
#include <ddraw.h>
#include <dsound.h>
#include <dmusici.h>
#include "gamelib.h"
#include "mygame.h"

/////////////////////////////////////////////////////////////////////////////
// CBall: Ball class
/////////////////////////////////////////////////////////////////////////////

CBall::CBall()
{
	is_alive = true;
	x = y = dx = dy = index = delay_counter = 0;
}

bool CBall::HitEraser(CEraser *eraser)
{
	// ˴lҺcxάO_Iy
	return HitRectangle(eraser->GetX1(), eraser->GetY1(),
		   eraser->GetX2(), eraser->GetY2());
}

bool CBall::HitRectangle(int tx1, int ty1, int tx2, int ty2)
{
	int x1 = x+dx;				// yWxy
	int y1 = y+dy;				// yWyy
	int x2 = x1 + bmp.Width();	// ykUxy
	int y2 = y1 + bmp.Height();	// ykUyy
	//
	// ˴yxλPѼƯxάO_涰
	//
	return (tx2 >= x1 && tx1 <= x2 && ty2 >= y1 && ty1 <= y2);
}

bool CBall::IsAlive()
{
	return is_alive;
}

void CBall::LoadBitmap()
{
	bmp.LoadBitmap(IDB_BALL,RGB(0,0,0));			// Jyϧ
	bmp_center.LoadBitmap(IDB_CENTER,RGB(0,0,0));	// Jyߪϧ
}

void CBall::Move()
{
	if (!is_alive)
		return;
	delay_counter--;
	if (delay_counter < 0) {
		delay_counter = delay;
		//
		// pyVߪ첾qdx, dy
		//
		const int STEPS=18;
		static const int DIFFX[]={35, 32, 26, 17, 6, -6, -17, -26, -32, -34, -32, -26, -17, -6, 6, 17, 26, 32, };
		static const int DIFFY[]={0, 11, 22, 30, 34, 34, 30, 22, 11, 0, -11, -22, -30, -34, -34, -30, -22, -11, };
		index++;
		if (index >= STEPS)
			index = 0;
		dx = DIFFX[index];
		dy = DIFFY[index];
	}
}

void CBall::SetDelay(int d)
{
	delay = d;
}

void CBall::SetIsAlive(bool alive)
{
	is_alive = alive;
}

void CBall::SetXY(int nx, int ny)
{
	x = nx; y = ny;
}

void CBall::Show()
{
	if (is_alive) {
		bmp.SetTopLeft(x+dx,y+dy);
		bmp.ShowBitmap();
		bmp_center.SetTopLeft(x,y);
		bmp_center.ShowBitmap();
	}
}

/////////////////////////////////////////////////////////////////////////////
// CEraser: Eraser class
/////////////////////////////////////////////////////////////////////////////

CEraser::CEraser()
{
	Initialize();
}

int CEraser::GetX1()
{
	return x;
}

int CEraser::GetY1()
{
	return y;
}

int CEraser::GetX2()
{
	return x + animation.Width();
}

int CEraser::GetY2()
{
	return y + animation.Height();
}

void CEraser::Initialize()
{
	const int X_POS = 280;
	const int Y_POS = 400;
	x = X_POS;
	y = Y_POS;
	isMovingLeft = isMovingRight= isMovingUp = isMovingDown = false;
}

void CEraser::LoadBitmap()
{
	animation.AddBitmap(IDB_ERASER1,RGB(255,255,255));
	animation.AddBitmap(IDB_ERASER2,RGB(255,255,255));
	animation.AddBitmap(IDB_ERASER3,RGB(255,255,255));
	animation.AddBitmap(IDB_ERASER2,RGB(255,255,255));
}

void CEraser::Move()
{
	const int STEP_SIZE = 2;
	animation.Move();
	if (isMovingLeft)
		x -= STEP_SIZE;
	if (isMovingRight)
		x += STEP_SIZE;
	if (isMovingUp)
		y -= STEP_SIZE;
	if (isMovingDown)
		y += STEP_SIZE;
}

void CEraser::SetMovingDown(bool flag)
{
	isMovingDown = flag;
}

void CEraser::SetMovingLeft(bool flag)
{
	isMovingLeft = flag;
}

void CEraser::SetMovingRight(bool flag)
{
	isMovingRight = flag;
}

void CEraser::SetMovingUp(bool flag)
{
	isMovingUp = flag;
}

void CEraser::SetXY(int nx, int ny)
{
	x = nx; y = ny;
}

void CEraser::Show()
{
	animation.SetTopLeft(x,y);
	animation.Show();
}

/////////////////////////////////////////////////////////////////////////////
// oclassCC}Ye
/////////////////////////////////////////////////////////////////////////////

CGameStateInit::CGameStateInit(CGame *g)
: CGameState(g)
{
}

void CGameStateInit::OnBeginState()
{
}

void CGameStateInit::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	const char KEY_ESC = 27;
	const char KEY_SPACE = ' ';
	if (nChar == KEY_SPACE)
		GotoGameState(GAME_STATE_RUN);	// GAME_STATE_RUN
	else if (nChar == KEY_ESC)			// Demo Ck
		PostQuitMessage(0);				// C
}

void CGameStateInit::OnLButtonDown(UINT nFlags, CPoint point)
{
	GotoGameState(GAME_STATE_RUN);		// GAME_STATE_RUN
}

void CGameStateInit::Show()
{
	//
	// DemoùrϥΡAL}oɽкɶqקKϥΦrACMovingBitmapn
	//
	CDC *pDC = CDDraw::GetBackCDC();			// o Back Plain  CDC 
	CFont f,*fp;
	f.CreatePointFont(160,"Times New Roman");	//  font f; 16016 pointr
	fp=pDC->SelectObject(&f);					//  font f
	pDC->SetBkColor(RGB(0,0,0));
	pDC->SetTextColor(RGB(255,255,0));
	pDC->TextOut(120,220,"Please click mouse or press SPACE to begin.");
	pDC->TextOut(5,395,"Press Ctrl-F to switch in between window mode and full screen mode.");
	pDC->TextOut(5,425,"Press Ctrl-Q to pause the Game.");
	pDC->TextOut(5,455,"Press Alt-F4 or ESC to Quit.");
	pDC->SelectObject(fp);						//  font f (dUn|F)
	CDDraw::ReleaseBackCDC();					//  Back Plain  CDC
}								

/////////////////////////////////////////////////////////////////////////////
// oclassCA(Game Over)
/////////////////////////////////////////////////////////////////////////////

CGameStateOver::CGameStateOver(CGame *g)
: CGameState(g)
{
}

void CGameStateOver::Move()
{
	counter--;
	if (counter < 0)
		GotoGameState(GAME_STATE_INIT);
}

void CGameStateOver::OnBeginState()
{
	counter = 30 * 5; // 5 seconds
}

void CGameStateOver::Show()
{
	CDC *pDC = CDDraw::GetBackCDC();			// o Back Plain  CDC 
	CFont f,*fp;
	f.CreatePointFont(160,"Times New Roman");	//  font f; 16016 pointr
	fp=pDC->SelectObject(&f);					//  font f
	pDC->SetBkColor(RGB(0,0,0));
	pDC->SetTextColor(RGB(255,255,0));
	char str[80];								// Demo Ʀrrꪺഫ
	sprintf(str, "Game Over ! (%d)", counter / 30);
	pDC->TextOut(240,210,str);
	pDC->SelectObject(fp);						//  font f (dUn|F)
	CDDraw::ReleaseBackCDC();					//  Back Plain  CDC
}

/////////////////////////////////////////////////////////////////////////////
// oclassCC檫ADnC{bo
/////////////////////////////////////////////////////////////////////////////

CGameStateRun::CGameStateRun(CGame *g)
: CGameState(g), NUMBALLS(28)
{
	ball = new CBall [NUMBALLS];
}

CGameStateRun::~CGameStateRun()
{
	delete [] ball;
}

void CGameStateRun::OnBeginState()
{
	const int BALL_GAP = 90;
	const int BALL_XY_OFFSET = 45;
	const int BALL_PER_ROW = 7;
	const int HITS_LEFT = 10;
	const int HITS_LEFT_X = 590;
	const int HITS_LEFT_Y = 0;
	const int BACKGROUND_X = 60;
	const int ANIMATION_SPEED = 15;
	for (int i = 0; i < NUMBALLS; i++) {				// ]wy_ly
		int x_pos = i % BALL_PER_ROW;
		int y_pos = i / BALL_PER_ROW;
		ball[i].SetXY(x_pos * BALL_GAP + BALL_XY_OFFSET, y_pos * BALL_GAP + BALL_XY_OFFSET);
		ball[i].SetDelay(x_pos);
		ball[i].SetIsAlive(true);
	}
	eraser.Initialize();
	background.SetTopLeft(BACKGROUND_X,0);				// ]wI_ly
	help.SetTopLeft(0, SIZE_Y - help.Height());			// ]wϪ_ly
	animation.SetTopLeft(-animation.Width(), SIZE_Y/2);	// ]wʵe_ly
	animation.SetDelayCount(ANIMATION_SPEED);			// ]wʵet
	hits_left.SetInteger(HITS_LEFT);					// wѤU
	hits_left.SetTopLeft(HITS_LEFT_X,HITS_LEFT_Y);		// wѤUƪy
	CAudio::Instance()->Play(AUDIO_LAKE, true);			//  WAVE
	CAudio::Instance()->Play(AUDIO_NTUT, true);			//  MIDI
}

void CGameStateRun::Move()								// ʹC
{
	//
	// pGƱקcursor˦AhNU{commmentYi
	//
	// SetCursor(AfxGetApp()->LoadCursor(IDC_GAMECURSOR));
	//
	// ʭIϪy
	//
	if (background.Top() > SIZE_Y)
		background.SetTopLeft(60 ,-background.Height());
	background.SetTopLeft(background.Left(),background.Top()+1);
	//
	// ʲy
	//
	int i;
	for (i=0; i < NUMBALLS; i++)
		ball[i].Move();
	//
	// l
	//
	eraser.Move();
	//
	// P_lO_Iy
	//
	for (i=0; i < NUMBALLS; i++)
		if (ball[i].IsAlive() && ball[i].HitEraser(&eraser)) {
			ball[i].SetIsAlive(false);
			CAudio::Instance()->Play(AUDIO_DING);
			hits_left.Add(-1);
			//
			// YѾlIƬ0AhGame OverA
			//
			if (hits_left.GetInteger() <= 0) {
				CAudio::Instance()->Stop(AUDIO_LAKE);	//  WAVE
				CAudio::Instance()->Stop(AUDIO_NTUT);	//  MIDI
				GotoGameState(GAME_STATE_OVER);
			}
		}
	//
	// ʰʵe
	//
	if (animation.Left() > SIZE_X)
		animation.SetTopLeft(-animation.Width(), SIZE_Y/2);
	animation.SetTopLeft(animation.Left()+1,animation.Top());
	animation.Move(); // ϱoanimation|bitmap
}

void CGameStateRun::OnInit()  								// CȤιϧγ]w
{
	for (int i = 0; i < NUMBALLS; i++)	
		ball[i].LoadBitmap();								// JiӲyϧ
	eraser.LoadBitmap();
	background.LoadBitmap(IDB_BACKGROUND);					// JIϧ
	help.LoadBitmap(IDB_HELP,RGB(255,255,255));				// Jϧ
	corner.LoadBitmap(IDB_CORNER);							// Jϧ
	corner.ShowBitmap(background);							// NcornerKbackground
	for (i = IDB_BALL1; i <= IDB_BALL4; i++)				// Jʵe(4iϧκc)
		animation.AddBitmap(i,RGB(0,0,0));
	hits_left.LoadBitmap();									
	CAudio::Instance()->Load(AUDIO_DING,  "sounds\\ding.wav");	// Js0nding.wav
	CAudio::Instance()->Load(AUDIO_LAKE,  "sounds\\lake.wav");	// Js1nlake.wav
	CAudio::Instance()->Load(AUDIO_NTUT,  "sounds\\ntut.mid");	// Js2nntut.mid
}

void CGameStateRun::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	const char KEY_LEFT  = 0x25; // keyboardbY
	const char KEY_UP    = 0x26; // keyboardWbY
	const char KEY_RIGHT = 0x27; // keyboardkbY
	const char KEY_DOWN  = 0x28; // keyboardUbY
	if (nChar == KEY_LEFT)
		eraser.SetMovingLeft(true);
	if (nChar == KEY_RIGHT)
		eraser.SetMovingRight(true);
	if (nChar == KEY_UP)
		eraser.SetMovingUp(true);
	if (nChar == KEY_DOWN)
		eraser.SetMovingDown(true);
}

void CGameStateRun::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	const char KEY_LEFT  = 0x25; // keyboardbY
	const char KEY_UP    = 0x26; // keyboardWbY
	const char KEY_RIGHT = 0x27; // keyboardkbY
	const char KEY_DOWN  = 0x28; // keyboardUbY
	if (nChar == KEY_LEFT)
		eraser.SetMovingLeft(false);
	if (nChar == KEY_RIGHT)
		eraser.SetMovingRight(false);
	if (nChar == KEY_UP)
		eraser.SetMovingUp(false);
	if (nChar == KEY_DOWN)
		eraser.SetMovingDown(false);
}

void CGameStateRun::OnLButtonDown(UINT nFlags, CPoint point)  // Bzƹʧ@
{
	eraser.SetMovingLeft(true);
}

void CGameStateRun::OnLButtonUp(UINT nFlags, CPoint point)	// Bzƹʧ@
{
	eraser.SetMovingLeft(false);
}

void CGameStateRun::OnRButtonDown(UINT nFlags, CPoint point)  // Bzƹʧ@
{
	eraser.SetMovingRight(true);
}

void CGameStateRun::OnRButtonUp(UINT nFlags, CPoint point)	// Bzƹʧ@
{
	eraser.SetMovingRight(false);
}

void CGameStateRun::Show()
{
	//
	//  `NGShow̭dUnʥ󪫥󪺮yСAʮyЪu@Move~A
	//        _hsøϮ(OnDraw)AN|ʡAݰ_ӷ|ܩǡCyNy
	//        AMovetdMVCModelAShowtdViewC
	//
	//
	//  KWIϡBơByBlBʵe
	//
	help.ShowBitmap();					// KW
	background.ShowBitmap();			// KWI
	hits_left.ShowBitmap();
	for (int i=0; i < NUMBALLS; i++)
		ball[i].Show();					// KWiy
	animation.Show();					// KWʵe
	eraser.Show();						// KWl
	//
	//  KWWΥkU
	//
	corner.SetTopLeft(0,0);
	corner.ShowBitmap();
	corner.SetTopLeft(SIZE_X-corner.Width(), SIZE_Y-corner.Height());
	corner.ShowBitmap();
}
